/*
 * Tribute to the old Atari ST demo scene
 *
 * This is a what a typical Atari ST demo screen could look like
 * around 1989; the fonts are from actual demos
 *
 * Done using Watcom C++ 11.0
 *
 * Mikael Kalms (1999-10-13), mikael@kalms.org
 */

#include <conio.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>


typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;


int mod2(int x, int y)
{
  return (x >= 0 ? x % y : (y - (-x % y)) % y);
}


byte vscreen[320 * 200];

byte font[320 * 200];
byte fontpal[16*3];
char *fontremap = "ABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890 !?,.'()-";

byte font2[320 * 200];
byte font2pal[16*3];

byte logo[320 * 70];

byte fuji[64 * 50];

byte scrollinternalbuf[320 * 4 * 25];
int scrollinternalpos = 0;


char *scrolltext = "   OLDSKOOL IS IN DA HOUSE!       ";

byte waterpal[] =
{ 0x00, 0x00, 0x00,
  0x00, 0x00, 0x24,
  0x00, 0x00, 0x48,
  0x00, 0x00, 0x6d,
  0x00, 0x00, 0x91,
  0x00, 0x00, 0xb6,
  0x00, 0x00, 0xda,
  0x00, 0x00, 0xff
};

void drawdistscroll(int frame)
{
  memset(&vscreen[50 * 320], 0, 100 * 320);
  int offs = mod2(scrollinternalpos + 160, 320 * 2);

  int xdist[25];
  int ydist[80];

  for (int dist = 0; dist < 25; dist++)
    xdist[dist] = sin(frame * 0.1 + dist * 0.15) * 7.9;

  for (dist = 0; dist < 80; dist++)
    ydist[dist] = sin(frame * 0.1 + dist * 0.03) * 15.9;

  byte *src, *dest;

  for (int x = 0; x < 80; x++)
    {
      src = &scrollinternalbuf[offs + x * 4];
      dest = &vscreen[(100 + ydist[x]) * 320 + x * 4];
      for (int y = 0; y < 25; y++)
	*((dword *) &dest[y * 320])
	  = *((dword *) &src[y * 320 * 4 + xdist[y]]);
    }
}

void drawwater(int frame)
{
  int watercols[] = 
  {
    16, 16, 16, 16,
    17, 16, 17, 17, 17,
    18, 17, 18, 18, 18,
    19, 18, 19, 19, 19, 19,
    20, 19, 20, 20, 20, 20,
    21, 20, 21, 21, 21, 21, 21,
    22, 21, 22, 22, 22, 22, 22, 22, 22,
    23, 22, 23, 23, 23, 23, 23, 23, 23,
    23, 23, 23, 23, 23, 23, 23,
    23, 23, 23, 23, 23, 23, 23,
  };
  for (int y = 0; y < 50; y++)
    {
      int bgcol = watercols[y];
      byte *src = &vscreen[(150 - (y + 1) * 2) * 320];
      byte *dest = &vscreen[(150 + y) * 320];
      int x = 320;
      do
	{
	  int col = *src++;
	  if (!col)
	    col = bgcol;
	  col += 16;
	  *dest++ = col;
	} while (--x);
    }
}

void drawlogo(int frame)
{
  int ydisp[50];
  for (int y = 0; y < 50; y++)
    ydisp[y] = sin(frame * 0.01 + y * 0.03) * sin(frame * 0.07 + y * 0.1) * 10;

  for (y = 0; y < 50; y++)
    {
      byte *src = &logo[(y + 10 + ydisp[y]) * 320 + 64];
      byte *dest = &vscreen[y * 320 + 64];
      memcpy(dest, src, 320 - 128);
    }
}

void drawfuji(int frame)
{
  for (int y = 0; y < 50; y++)
    {
      memcpy(&vscreen[y * 320], &fuji[y * 64], 64);
      memcpy(&vscreen[y * 320 + 256], &fuji[y * 64], 64);
    }
}

void setfujipalette(int frame)
{
  int nentries = 8*8;
  int palette[] =
  { 0x777, 0x777, 0x776, 0x776, 0x775, 0x775, 0x774, 0x774,
    0x773, 0x773, 0x772, 0x772, 0x771, 0x771, 0x770, 0x770,
    0x770, 0x770, 0x671, 0x671, 0x572, 0x572, 0x473, 0x473,
    0x374, 0x374, 0x275, 0x275, 0x176, 0x176, 0x077, 0x077,
    0x077, 0x077, 0x167, 0x167, 0x257, 0x257, 0x347, 0x347,
    0x437, 0x437, 0x527, 0x527, 0x617, 0x617, 0x707, 0x707,
    0x707, 0x707, 0x717, 0x717, 0x727, 0x727, 0x737, 0x737,
    0x747, 0x747, 0x757, 0x757, 0x767, 0x767, 0x777, 0x777,
    0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777,
    0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777,
    0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777, 0x777,
    0
    };

  int palidx = mod2(frame / 3, nentries);
  for (int i = 0; i < 50; i++)
    {
      outp(0x3c8, 128 + i);
      int col = (palette[palidx] & 0x700) >> 5;
      col |= col >> 3;
      outp(0x3c9, col);
      col = (palette[palidx] & 0x70) >> 1;
      col |= col >> 3;
      outp(0x3c9, col);
      col = (palette[palidx] & 0x7) << 3;
      col |= col >> 3;
      outp(0x3c9, col);
      palidx++;
      palidx = mod2(palidx, nentries);
    }
}

void scrollblitchar(int letter, int blk)
{
  int xoffs = (letter % 10) * 32;
  int yoffs = (letter / 10) * 25;
  for (int y = 0; y < 25; y++)
    memcpy(&scrollinternalbuf[y * 320 * 4 + blk * 32],
	   &font[(y + yoffs) * 320 + xoffs],
	   32);
}

void scrollresetwindow(void)
{
  scrollinternalpos = -40*32;
}

void scrollsetwindow(char *scrolltext, int pos)
{
  int textlen = strlen(scrolltext);

  int currblk = (pos / 32);
  int prevblk = (scrollinternalpos / 32);

  int delta = 0;
  if (currblk > prevblk)
    {
      currblk += (320 * 2 / 32);
      prevblk += (320 * 2 / 32);
      delta = 1;
    }
  else if (currblk < prevblk)
    delta = -1;

  while (currblk != prevblk)
    {
      if (delta < 0)
	prevblk += delta;
      int textpos = mod2(prevblk, textlen);
      char *hit = strchr(fontremap, toupper(scrolltext[textpos]));
      if (!hit)
	hit = strchr(fontremap, ' ');
      int letter = hit - fontremap;
      int blk = mod2(prevblk, 320 * 2 / 32);
      scrollblitchar(letter, blk);
      scrollblitchar(letter, blk + 320 * 2 / 32);
      if (delta > 0)
	prevblk += delta;
    }

  scrollinternalpos = pos;
}

bool loadpc1(char *filename, byte *outputpic, byte *outputpal)
{
  // Load source file into buffer
  FILE *f;
  if (!(f = fopen(filename, "rb")))
    return false;
  fseek(f, 0, SEEK_END);
  int sourcelen = ftell(f);
  byte *sourcebuf = new byte[sourcelen];
  fseek(f, 0, SEEK_SET);
  if (!(fread(sourcebuf, sourcelen, 1, f)))
    {
      fclose(f);
      return false;
    }
  fclose(f);

  // Verify .pc1 format
  if (sourcebuf[0] != 0x80 || sourcebuf[1] != 0x00)
    {
      delete [] sourcebuf;
      return false;
    }

  byte *srcptr = &sourcebuf[2];

  // Extract palette
  for (int col = 0; col < 16; col++)
    {
      byte c0 = *srcptr++;
      byte c1 = *srcptr++;
      outputpal[col * 3] = ((c0 & 0x07) << 5) | ((c0 & 0x07) << 2);
      outputpal[col * 3 + 1] = ((c1 & 0x70) << 1) | ((c1 & 0x70) >> 2);
      outputpal[col * 3 + 2] = ((c1 & 0x07) << 5) | ((c1 & 0x07) << 2);
    }


  // Decrunch image
  byte *destptr = outputpic;

  for (int y = 0; y < 200; y++)
    {
      fflush(stdout);
      byte bpls[4][40];

      // Decrunch line
      for (int bpl = 0; bpl < 4; bpl++)
	{
	  // Decrunch bpl
	  int outputcount = 0;
	  while (outputcount < 40)
	    {
	      int count = (signed char) *srcptr++;
	      if (count >= 0)
		{
		  while (count-- >= 0)
		    bpls[bpl][outputcount++] = *srcptr++;
		}
	      else if (count > -128)
		{
		  byte srcbyte = *srcptr++;
		  while (count++ <= 0)
		    bpls[bpl][outputcount++] = srcbyte;
		}
	    }
	}

      // p2c line
      for (int bytecount = 0; bytecount < 40; bytecount++)
	{
	  byte bpl0 = bpls[0][bytecount];
	  byte bpl1 = bpls[1][bytecount];
	  byte bpl2 = bpls[2][bytecount];
	  byte bpl3 = bpls[3][bytecount];
	  for (int bitcount = 0; bitcount < 8; bitcount++)
	    {
	      *destptr++ = (((bpl0 & 0x80) >> 7) | ((bpl1 & 0x80) >> 6)
			    | ((bpl2 & 0x80) >> 5) | ((bpl3 & 0x80) >> 4));
	      bpl0 <<= 1;
	      bpl1 <<= 1;
	      bpl2 <<= 1;
	      bpl3 <<= 1;
	    }
	}
    }

  delete [] sourcebuf;
  return true;
}

void vsync(void)
{
  __asm
    {
      pusha
      mov dx,0x3da;
    l1:
      in al,dx;
      test al,8;
      jz l1;
    l2:
      in al,dx;
      test al,8;
      jnz l2;
      popa
    }
}


void clearvscreen(void)
{
  memset(vscreen, 0, 320 * 200);
}


void updatescreen(void)
{
  memcpy((char *) 0xa0000, vscreen, 320 * 200);
}


void setgfxmode(int mode)
{
  __asm
    {
      mov eax,mode;
      int 10h;
    }
}


void setpal(byte *pal, int start, int count)
{
  outp(0x3c8, start);
  for (int i = 0; i < count * 3; i++)
    outp(0x3c9, pal[i] >> 2);
}

void setcol(int col, int rgb)
{
  outp(0x3c8, col);
  outp(0x3c9, (rgb >> 16) & 0xff);
  outp(0x3c9, (rgb >> 8) & 0xff);
  outp(0x3c9, rgb & 0xff);
}

void rectblit(byte *dest, byte *src, int x1, int y1, int x0, int y0,
	     int w, int h)
{
  for (int height = 0; height < h; height++)
    memcpy(dest + (y1 + height) * 320 + x1, src + (y0 + height) * 320 + x0, w);
}

int init(void)
{
  if (!loadpc1("font1.pc1", font, fontpal))
    return false;

  if (!loadpc1("font2.pc1", font2, font2pal))
    return false;

  for (int i = 0; i < 320 * 200; i++)
    font2[i] += 64;

  FILE *f;
  if (!(f = fopen("fuji.raw", "rb")))
    return false;
  fread(fuji, 64*50, 1, f);
  fclose(f);

  for (i = 0; i < 50; i++)
    {
      for (int j = 0; j < 64; j++)
	if (fuji[i * 64 + j])
	  fuji[i * 64 + j] = 128 + i;
    }

  rectblit(logo, font2, 160-64, 20, 32*1, 32*0, 32, 32);
  rectblit(logo, font2, 160-32, 20, 32*1, 32*1, 32, 32);
  rectblit(logo, font2, 160, 20, 32*0, 32*0, 32, 32);
  rectblit(logo, font2, 160+32, 20, 32*7, 32*0, 32, 32);

  return true;
}

int main(void)
{
  if (!init())
    return 1;

  setgfxmode(0x13);
  setpal(fontpal, 0, 16);

  byte fontpal2[16*3];
  for (int i = 0; i < 16*3; i++)
    {
      int col = fontpal[i] >> 6;
      fontpal2[i] = (col << 5) | (col << 2);
    }
  setpal(fontpal2, 16, 16);
  setpal(waterpal, 32, 8);

  setpal(font2pal, 64, 16);
  setfujipalette(0);

  float t = 0;
  scrollresetwindow();
  int scrollpos = 0;
  while (!kbhit())
    {
      scrollpos = t * 4 - 10*32;
      scrollsetwindow(scrolltext, scrollpos);
      drawfuji(t);
      drawlogo(t);
      drawdistscroll(t);
      drawwater(t);
      //setcol(0, 0x000000);
      vsync();
      //setcol(0, 0xff0000);
      setfujipalette(t);
      updatescreen();
      t++;
    }
  setgfxmode(3);
  return 0;
}
