/*
 * Image viewer for *.PC1 (Degas Elite Compressed) format images
 * [Degas Elite is a paint program for the Atari ST]
 *
 * Done using Watcom C++ 11.0
 *
 * Mikael Kalms (1999-10-13), mikael@kalms.org
 */

#include <conio.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;

byte image[320 * 200];
byte imagepal[16*3];

bool loadpc1(char *filename, byte *outputpic, byte *outputpal)
{
  // Load source file into buffer
  FILE *f;
  if (!(f = fopen(filename, "rb")))
    return false;
  fseek(f, 0, SEEK_END);
  int sourcelen = ftell(f);
  byte *sourcebuf = new byte[sourcelen];
  fseek(f, 0, SEEK_SET);
  if (!(fread(sourcebuf, sourcelen, 1, f)))
    {
      fclose(f);
      return false;
    }
  fclose(f);

  // Verify .pc1 format
  if (sourcebuf[0] != 0x80 || sourcebuf[1] != 0x00)
    {
      delete [] sourcebuf;
      return false;
    }

  byte *srcptr = &sourcebuf[2];

  // Extract palette
  for (int col = 0; col < 16; col++)
    {
      byte c0 = *srcptr++;
      byte c1 = *srcptr++;
      outputpal[col * 3] = ((c0 & 0x07) << 5) | ((c0 & 0x07) << 2);
      outputpal[col * 3 + 1] = ((c1 & 0x70) << 1) | ((c1 & 0x70) >> 2);
      outputpal[col * 3 + 2] = ((c1 & 0x07) << 5) | ((c1 & 0x07) << 2);
    }


  // Decrunch image
  byte *destptr = outputpic;

  for (int y = 0; y < 200; y++)
    {
      fflush(stdout);
      byte bpls[4][40];

      // Decrunch line
      for (int bpl = 0; bpl < 4; bpl++)
	{
	  // Decrunch bpl
	  int outputcount = 0;
	  while (outputcount < 40)
	    {
	      int count = (signed char) *srcptr++;
	      if (count >= 0)
		{
		  while (count-- >= 0)
		    bpls[bpl][outputcount++] = *srcptr++;
		}
	      else if (count > -128)
		{
		  byte srcbyte = *srcptr++;
		  while (count++ <= 0)
		    bpls[bpl][outputcount++] = srcbyte;
		}
	    }
	}

      // p2c line
      for (int bytecount = 0; bytecount < 40; bytecount++)
	{
	  byte bpl0 = bpls[0][bytecount];
	  byte bpl1 = bpls[1][bytecount];
	  byte bpl2 = bpls[2][bytecount];
	  byte bpl3 = bpls[3][bytecount];
	  for (int bitcount = 0; bitcount < 8; bitcount++)
	    {
	      *destptr++ = (((bpl0 & 0x80) >> 7) | ((bpl1 & 0x80) >> 6)
			    | ((bpl2 & 0x80) >> 5) | ((bpl3 & 0x80) >> 4));
	      bpl0 <<= 1;
	      bpl1 <<= 1;
	      bpl2 <<= 1;
	      bpl3 <<= 1;
	    }
	}
    }

  delete [] sourcebuf;
  return true;
}

void setgfxmode(int mode)
{
  __asm
    {
      mov eax,mode;
      int 10h;
    }
}


void setpal(byte *pal, int start, int count)
{
  outp(0x3c8, start);
  for (int i = 0; i < count * 3; i++)
    outp(0x3c9, pal[i] >> 2);
}

int main(int argc, char **argv)
{
  if (argc != 2)
    {
      printf("Usage: SHOWPC1 <image>\n");
      return 0;
    }

  if (!loadpc1(argv[1], image, imagepal))
    return 0;

  setgfxmode(0x13);
  setpal(imagepal, 0, 16);
  memcpy((byte *) 0xa0000, image, 320 * 200);
  getch();
  setgfxmode(3);
  return 0;
}
